//
// Copyright 2007 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include <fltk/Window.h>
#include <fltk/ScrollGroup.h>
#include <fltk/FileChooser.h>
#include <fltk/file_chooser.h>
#include <fltk/Box.h>
#include <fltk/Style.h>
#include <fltk/Input.h>
#include <fltk/IntInput.h>
#include <fltk/MenuBar.h>
#include <fltk/Browser.h>
#include <fltk/Item.h>
#include <fltk/ColorChooser.h>
#include <fltk/ask.h>
#include <fltk/events.h>
#include <fltk/utf.h>
#include <fltk/run.h>

#include "PSEditor.H"
#include "util.h"
#include "../config.h"

PSEditor *psed_p   = NULL;
fltk::ScrollGroup *scroll = NULL;
fltk::Browser *page_sel = NULL;

int xev_handler(int ev, Window *w) {
	if (psed_p) {
		return psed_p->handleX11(ev);
	} else {
		return 0;
	}
}

void save_cb();

int check_save(void) {
	if (!psed_p->modified()) return 1;

	int r = fltk::choice("The current file has not been saved.\n"
		"Would you like to save it now?",
		"Cancel", "Save", "Discard Changes");

	if (r == 1) {
		save_cb(); // Save the file...
		return !psed_p->modified();
	}

	return (r == 2) ? 1 : 0;
}

static int
confirm_overwrite(const char *f) {
	struct stat sb;
	if (stat(f, &sb) == 0) {
		return fltk::choice("The file exists.\n", "Cancel", "Overwrite", NULL);
	} else {
		return 1;
	}
}

char filename[256] = "";


void page_sel_cb(fltk::Widget *w, void *) {
	int p = page_sel->value() + 1;
	if (p > 0) {
		psed_p->load_page(p);
	}
}

void page_sel_fill() {
	char buf[64];
	int p = psed_p->get_pages();

	page_sel->clear();

	if (p == 0) {
		return;
	}

	for(int i=1; i<=p; i++) {
		snprintf(buf, sizeof(buf), "%d", i);
		page_sel->add(buf);
	}
	page_sel->select(psed_p->get_page() - 1);
}

void open_file(const char *file) {
	char *loc_file;
	FILE *p;
	int status;
	char *args[32];
	pid_t pid;
	char *dot;

	loc_file = (char*) malloc(strlen(file) * 3);
	utf8tomb(file, strlen(file), loc_file, strlen(file) * 3);

	dot = strrchr(loc_file, '.');
	if (dot && strcasecmp(dot, ".pdf") == 0) {
		// assume pdf and try to import it using pdftops
		args[0] = "pdftops";
		args[1] = loc_file;
		args[2] = "-";
		args[3] = NULL;

		p = pexecvp("pdftops", args, &pid, "r");

		if (p) {
			psed_p->open_file(p);
			psed_p->load();
			fclose(p);
			waitpid(pid, &status, 0);
			if (WEXITSTATUS(status) == 127 || WEXITSTATUS(status) == 126) {
				fltk::message("PDF import depends on pdftops from xpdf.\n"
					"Make sure pdftops is available on your system.\n");
			} else if (WEXITSTATUS(status) != 0) {
				fltk::message("PDF import failed\n");
			}
		} else {
			perror("pexecvp");
		}
	} else {
		// assume postscript
		psed_p->open_file(loc_file);
		psed_p->load();
	}

	free(loc_file);
	page_sel_fill();
}

void open_cb() {
	if (!check_save()) return;

	const char *file = fltk::file_chooser("Open File?", "*.ps", filename);

	if (file != NULL) {
		open_file(file);
	}  
}

void export_pdf_cb() {
	const char *file;
	FILE *p;
	int status;
	char *args[32];
	pid_t pid;
	char *loc_file;

	file = fltk::file_chooser("Export as PDF to File?", "*.pdf", filename);

	if (file != NULL && confirm_overwrite(file)) {
		loc_file = (char*) malloc(strlen(file) * 3);
		utf8tomb(file, strlen(file), loc_file, strlen(file) * 3);

		args[0] = "ps2pdf";
		args[1] = "-";
		args[2] = loc_file;
		args[3] = NULL;

		signal(SIGPIPE, SIG_IGN);

		p = pexecvp("ps2pdf", args, &pid, "w");

		if (p) {
			psed_p->save(p);

			fclose(p);
			waitpid(pid, &status, 0); 
			if (WEXITSTATUS(status) == 127 || WEXITSTATUS(status) == 126) {
				fltk::message("PDF export depends on ps2pdf from ghostscript.\n"
					"Make sure ps2pdf is available on your system.\n");
			} else if (WEXITSTATUS(status) != 0) {
				fltk::message("PDF export failed\n");
			}
		} else {
			perror("pexecvp");
		}

		free(loc_file);
		signal(SIGPIPE, SIG_DFL);
	}
}

void import_cb() {
	const char *file;
	char *loc_file;

	file = fltk::file_chooser("Import Overlay from File?", "*.ps", filename);

	if (file != NULL) {
		loc_file = (char*) malloc(strlen(file) * 3);
		utf8tomb(file, strlen(file), loc_file, strlen(file) * 3);

		psed_p->import(loc_file);

		free(loc_file);
	}  
}

void first_cb() {
	psed_p->load();
	page_sel->select(psed_p->get_page() - 1);
}

void next_cb() {
	psed_p->next();
	page_sel->select(psed_p->get_page() - 1);
}

void quit_cb() {
	if (!check_save()) return;
	delete psed_p;
	exit(0);
}

void save_cb() {
	const char *file;
	char *loc_file;

	file = fltk::file_chooser("Save to File?", "*.ps", filename);

	if (file != NULL && confirm_overwrite(file)) {
		loc_file = (char*) malloc(strlen(file) * 3);
		utf8tomb(file, strlen(file), loc_file, strlen(file) * 3);

		psed_p->save(loc_file);

		free(loc_file);
	}
}

void print_cb() {
	char tmpname[256];
	char buf[256];
	int tmp_fd;

	int r = fltk::choice("Print file?", "Cancel", "Print", NULL);
	if (r != 1) {
		return;
	}

	strncpy(tmpname, "/tmp/PSEditWidgetXXXXXX", 256);
	tmp_fd = mkstemp(tmpname);

	if (tmp_fd >= 0) {
		close(tmp_fd);
		if (psed_p->save(tmpname) != 0) {
			fprintf(stderr, "Failed to print file\n");
		} else {
			snprintf(buf, 256, "lpr %s", tmpname);
			system(buf);
		}
		unlink(tmpname);
	} 
}

void about_cb() {
	fltk::message("flpsed -- a PostScript annotator\n"
		"Version %s\n\n"
		"(c) Johannes Hofmann 2004-2007\n\n"
		"PostScript is a registered trademark of Adobe Systems", VERSION);
}

fltk::Choice *size_c;
fltk::Button *color_b;

static struct {
	char *label;
	int size;
} text_sizes[] = {
	{"8", 8},
	{"10", 10},
	{"12", 12},
	{"14", 14},
	{"18", 18},
	{"24", 24},
	{NULL, 0}
};

void property_changed_cb() {
	PSEditColor c;
	int size;

	psed_p->get_color(&c);

	// XXX No idea why color(0,0,0) gives grey
	if (c.get_r() + c.get_g() + c.get_b() == 0) {
		color_b->color(BLACK);
	} else {    
		color_b->color(fltk::color(c.get_r(), c.get_g(), c.get_b()));
	}
	color_b->redraw();

	size = psed_p->get_size();
	for (int i=0; text_sizes[i].label != NULL; i++) {
		if (size == text_sizes[i].size) {
			size_c->value(i);
			size_c->redraw();
		}
	}
}

void size_cb(fltk::Widget *w, void *) {
	fltk::Menu* mw = (fltk::Menu*)w;
	fltk::Widget* m = mw->item();
	if (m) {
		psed_p->set_size(atoi(m->label()));
	}
}

void color_cb(fltk::Widget *w, void *v) {
	uchar r, g, b;
	PSEditColor pc;

	psed_p->get_color(&pc);

	r = pc.get_r();
	g = pc.get_g();
	b = pc.get_b();

	if (!fltk::color_chooser("Text Color", r, g, b)) return;
	fltk::Button* button = (fltk::Button*)v;
	pc.set(r, g, b);
	psed_p->set_color(&pc);
	button->color(fltk::color(r, g, b));
	button->parent()->redraw();
}

void zoom_cb(fltk::Widget *w, void *) {
	fltk::Menu* mw = (fltk::Menu*)w;
	if (mw) {
		if (psed_p) {
			psed_p->zoom(atoi(mw->label()));
		}
		if (scroll) {
			scroll->init_sizes();
			scroll->scrollTo(0,0);
			scroll->redraw();
		}
	}
}

void show_tags_cb(fltk::Widget* w, void*d) {
	fltk::Menu* mw = (fltk::Menu *)w;
	if (mw) {
		psed_p->set_show_tags(mw->flags() & STATE);
	}
}

void edit_tag_cb() {
	char *tag = psed_p->get_tag();
	const char *new_tag;
	new_tag = fltk::input("Tag Name", tag?tag:"");
	if (new_tag) {
		if (strcmp(new_tag, "") != 0) {
			psed_p->set_tag(new_tag);
		} else {
			psed_p->set_tag(NULL);
		}
	}
}

void fill_sizechoice(Choice *c) {
	c->begin();
	for (int i=0; i<6; i++) {
		c->add(text_sizes[i].label,  0, (Callback*)size_cb);
	}
	c->end();
}

void fill_menubar(MenuBar* mb) {
	mb->begin();
	mb->add("&File/&Open File...", CTRL+'o', (Callback*)open_cb);
	mb->add("&File/&Save File as...", CTRL+'s', (Callback*)save_cb);
	mb->add("&File/I&mport Tags from File...", CTRL+'m', (Callback*)import_cb);
	mb->add("&File/I&mport PDF...", CTRL+'i', (Callback*)open_cb);
	mb->add("&File/E&xport PDF...", CTRL+'x', (Callback*)export_pdf_cb);
	mb->add("&File/&Print", CTRL+'p', (Callback*)print_cb);
	mb->add("&File/&Quit", CTRL+'q', (Callback*)quit_cb);

	mb->add("&Page/&First", CTRL+'f', (Callback*)first_cb);
	mb->add("&Page/&Next", CTRL+'n', (Callback*)next_cb);

	mb->add("&Zoom/50 %", 0, (Callback*)zoom_cb);
	mb->add("&Zoom/75 %", 0, (Callback*)zoom_cb);
	mb->add("&Zoom/100 %", 0, (Callback*)zoom_cb);
	mb->add("&Zoom/150 %", 0, (Callback*)zoom_cb);
	mb->add("&Zoom/200 %", 0, (Callback*)zoom_cb);
	mb->add("&Zoom/250 %", 0, (Callback*)zoom_cb);

	mb->add("&Tags/Show &Tags", CTRL+'t', (Callback*)show_tags_cb, (void*) 0, fltk::MENU_TOGGLE|STATE);
	mb->add("&Tags/&Edit Tag", CTRL+'e', (Callback*)edit_tag_cb);


	mb->add("&Help/About", 0, (Callback*)about_cb);
	mb->end();
}

void usage() {
	fprintf(stderr,
		"usage: flpsed [-hbdz] [-t <tag>=<value>] [<infile>] [<outfile>]\n"
		"   -h                 print this message\n"
		"   -b                 batch mode (no gui)\n"
		"   -d                 dump tags and values from a document\n"
		"                      to stdout (this implies -b)\n"
		"   -z <zoom>          set the zoom percentage (e.g. 200)\n"
		"   -t <tag>=<value>   set text to <value> where tag is <tag>\n"
		"   <infile>           optional input file; in batch mode if no\n"
		"                      input file is given, input is read from stdin\n"
		"   <outfile>          optional output file for batch mode; if no\n"
		"                      output file is given, output is written to stdout\n\n"
		"additionally flpsed supports the following FLTK standard options:\n"
		" -geometry WxH+X+Y\n"
		" -iconic\n"
		" -scheme string\n"
	);
}

#define TV_LEN 256

int main(int argc, char** argv) {
	char *sep, *tmp, **my_argv;
	int c, err, bflag = 0, dflag = 0;
	int i, n;
	fltk::Window *win;
	fltk::MenuBar *m;
	struct {char *tag; char *value;} tv[TV_LEN];
	int tv_idx = 0, my_argc;
	FILE *out_fp = NULL;
	int zoom_val = 0;

	err = 0;
	while ((c = getopt(argc, argv, "hdbt:z:g:d:f:i:s:")) != -1) {
		switch (c) {  
			case 'h':
				usage();
				exit(0);
				break;
			case 'b':
				bflag = 1;
				break;
			case 'd':
				dflag = 1;
				break;
			case 't':
				tmp = strdup(optarg);
				sep = strchr(tmp, '=');
				if (!sep) {
					fprintf(stderr, "Cannot parse %s\n", optarg);
					free(tmp);
					continue;
				}
				*sep = '\0';

				if (tv_idx >= TV_LEN) {
					fprintf(stderr, "More than %d tag/value pairs; ignoring %s->%s\n", 
						TV_LEN, tmp, sep+1);
				} else {
					tv[tv_idx].tag = strdup(tmp);
					tv[tv_idx].value = strdup(sep+1);
					tv_idx++;
				}
				free(tmp);
				break;
			case 'z':
				zoom_val = atoi(optarg);
				break;
			default:
				i = optind -1;
				n = fltk::arg(argc, argv, i);
				if (n == 0) {
					err++;
				} else {
					optind = i;
				}
		}
	}

	if (err) {
		usage();
		exit(1);
	}

	my_argc = argc - optind;
	my_argv = argv + optind;

	if (bflag || dflag) {
		//
		// Batch Mode 
		//

		PSEditModel *m = new PSEditModel();
		int tmp_fd;
		FILE *in_fp;

		if (my_argc >= 1) {
			in_fp = fopen(my_argv[0], "r");
			if (!in_fp) {
				perror("fopen");
				exit(1);
			}
		} else {
			in_fp = stdin;
		}


		tmp_fd= m->load(in_fp);

		if (tmp_fd == -1) {
			fprintf(stderr, "Could not load %s\n", argv[argc - 2]);
			exit(1);
		}

		if (in_fp != stdin) {
			fclose(in_fp);
		}

		for(int i=0; i<tv_idx; i++) {
			m->replace_tag(tv[i].tag, tv[i].value);
			free(tv[i].tag);
			free(tv[i].value);
		}


		if (bflag) {
			if (my_argc >= 2) {
				out_fp = fopen(my_argv[1], "w");
				if (!in_fp) {
					perror("fopen");
					exit(1);
				}
			} else {
				out_fp = stdout;
			}

			m->save(out_fp, tmp_fd);

			if (out_fp != stdout) {
				fclose(out_fp);
			}
		} else { // dump tags
			m->dump_tags();
		}

	} else {
		// 
		// Interactive Mode
		//

		win = new fltk::Window(600,700);
		win->begin();
		m = new fltk::MenuBar(0, 0, 600, 30);
		fill_menubar(m); 

		fltk::Group props(0, 30, 600, 25);
		props.box(fltk::ENGRAVED_BOX);
		props.begin();
		props.resizable(NULL);
		size_c = new fltk::Choice(30, 2, 50, 21, NULL);
		size_c->callback(size_cb);
		fill_sizechoice(size_c);
		size_c->tooltip("Text Size");

		color_b = new fltk::Button(100, 2, 21, 21);
		color_b->color(BLACK);
		color_b->callback(color_cb, color_b);
		color_b->shortcut(ALT + 'c');
		color_b->tooltip("Text Color");
		props.end();
		page_sel = new Browser(0, 55, 40, win->h()-55);
		page_sel->callback(page_sel_cb);
		page_sel->end();

		scroll = new fltk::ScrollGroup(40, 55, win->w() - 40, win->h() - 55);
		scroll->begin();

		psed_p = new PSEditor(0, 0, 700, 900);
		psed_p->property_changed_callback(property_changed_cb);
		scroll->end();

		fltk::open_display();
		fltk::add_event_handler(xev_handler);

		win->resizable(scroll);

		win->end();
		win->callback((fltk::Callback *)quit_cb);
		win->show(1, argv); 


		if (zoom_val) {
			psed_p->zoom(zoom_val);
		}

		if (my_argc >= 1) {
			open_file(my_argv[0]);
		}

		for(int i=0; i<tv_idx; i++) {
			psed_p->replace_tag(tv[i].tag, tv[i].value);
			free(tv[i].tag);
			free(tv[i].value);
		}

		return fltk::run();
	}
}
